using System.Numerics;
using Frontend.Assets;
using Frontend.Marketplaces.Interfaces;
using Nethereum.Contracts.Standards.ERC20.ContractDefinition;
using Nethereum.Web3;
using Nethereum.Web3.Accounts;
using Newtonsoft.Json.Linq;

namespace Frontend.Marketplaces;

public class PotatoMarketplaceMock(
    Account marketplaceAccount,
    Web3 marketplaceWeb3Client,
    string wethAddress
) : IPotatoMarketPlace
{
    private const double potatoKgChfPrice = 2.02; // source: https://www.blw.admin.ch/blw/de/home/markt/marktbeobachtung/kartoffeln.html

    private readonly Account account = marketplaceAccount;
    private readonly Web3 web3 = marketplaceWeb3Client;
    private readonly string wethAddress = wethAddress;

    public Task<double> EthPriceInChf => EthChfPrice();
    public Task<double> PotatoKgPriceInChf => Task.FromResult(potatoKgChfPrice);
    public Task<double> PotatoKgPriceInEth => PotatoKgEthPrice();

    public string Address => account.Address;

    public async Task<PotatoTransaction?> Buy(string ownerAddress)
    {
        var allowanceHandler = web3.Eth.GetContractQueryHandler<AllowanceFunction>();
        var allowanceFunction = new AllowanceFunction
        {
            Owner = ownerAddress,
            Spender = account.Address
        };

        var allowance = await allowanceHandler.QueryAsync<BigInteger>(wethAddress, allowanceFunction);

        var transferHandler = web3.Eth.GetContractTransactionHandler<TransferFromFunction>();
        var transferFunction = new TransferFromFunction
        {
            From = ownerAddress,
            To = account.Address,
            Value = allowance,
        };

        var receipt = await transferHandler.SendRequestAndWaitForReceiptAsync(wethAddress, transferFunction);
        if (receipt.Status.Value == 0)
        {
            return null;
        }

        var potatoWeight = (double)allowance / (double)await PotatoKgWeiPrice();
        return new PotatoTransaction
        {
            Weight = potatoWeight
        };
    }

    public async Task<bool> Sell(PotatoTransaction transaction, string receiverAddress)
    {
        var amount = (double)await PotatoKgWeiPrice() * transaction.Weight;

        var transferHandler = web3.Eth.GetContractTransactionHandler<TransferFunction>();
        var transferFunction = new TransferFunction
        {
            To = receiverAddress,
            Value = new BigInteger(amount),
        };

        var receipt = await transferHandler.SendRequestAndWaitForReceiptAsync(wethAddress, transferFunction);
        return receipt.Status.Value == 1;
    }

    private static async Task<double> EthChfPrice()
    {
        string url = "https://min-api.cryptocompare.com/data/price?fsym=ETH&tsyms=CHF";

        using var client = new HttpClient();
        HttpResponseMessage response = await client.GetAsync(url);
        response.EnsureSuccessStatusCode();
        string responseBody = await response.Content.ReadAsStringAsync();
        JObject json = JObject.Parse(responseBody);
        double priceInChf = json["CHF"]!.Value<double>();
        return priceInChf;
    }

    private static async Task<double> PotatoKgEthPrice()
    {
        var chfEthPrice = 1 / await EthChfPrice();
        return chfEthPrice * potatoKgChfPrice;
    }

    private static async Task<BigInteger> PotatoKgWeiPrice()
    {
        return Web3.Convert.ToWei(await PotatoKgEthPrice());
    }
}
